#ifndef _MC68328DIGI_H
#define _MC68328DIGI_H

#include <linux/ioctl.h>

#define PORT_A_DIR  (*(volatile unsigned char*)0xfffff400)
#define PORT_A_DAT  (*(volatile unsigned char*)0xfffff401)
#define PORT_A_PUL  (*(volatile unsigned char*)0xfffff402)
#define PORT_A_SEL  (*(volatile unsigned char*)0xfffff403)

#define PORT_B_DIR  (*(volatile unsigned char*)0xfffff408)
#define PORT_B_DAT  (*(volatile unsigned char*)0xfffff409)
#define PORT_B_PUL  (*(volatile unsigned char*)0xfffff40a)
#define PORT_B_SEL  (*(volatile unsigned char*)0xfffff40b)

#define PORT_C_DIR  (*(volatile unsigned char*)0xfffff410)
#define PORT_C_DAT  (*(volatile unsigned char*)0xfffff411)
#define PORT_C_PUL  (*(volatile unsigned char*)0xfffff412)
#define PORT_C_SEL  (*(volatile unsigned char*)0xfffff413)


#define PORT_D_DIR  (*(volatile unsigned char*)0xfffff418)
#define PORT_D_DAT  (*(volatile unsigned char*)0xfffff419)
#define PORT_D_PUL  (*(volatile unsigned char*)0xfffff41a)
#define PORT_D_SEL  (*(volatile unsigned char*)0xfffff41b)
#define PORT_D_POL  (*(volatile unsigned char*)0xfffff41c)
#define PORT_D_IRQEN  (*(volatile unsigned char*)0xfffff41d)
#define PORT_D_KBEN  (*(volatile unsigned char*)0xfffff41e)
#define PORT_D_IQEN (*(volatile unsigned char*)0xfffff41f)

#define PORT_E_DIR  (*(volatile unsigned char*)0xfffff420)
#define PORT_E_DAT  (*(volatile unsigned char*)0xfffff421)
#define PORT_E_PUL  (*(volatile unsigned char*)0xfffff422)
#define PORT_E_SEL  (*(volatile unsigned char*)0xfffff423)

#define PORT_F_DIR  (*(volatile unsigned char*)0xfffff428)
#define PORT_F_DAT  (*(volatile unsigned char*)0xfffff429)
#define PORT_F_PUL  (*(volatile unsigned char*)0xfffff42a)
#define PORT_F_SEL  (*(volatile unsigned char*)0xfffff42b)

#define PORT_G_DIR  (*(volatile unsigned char*)0xfffff430)
#define PORT_G_DAT  (*(volatile unsigned char*)0xfffff431)
#define PORT_G_PUL  (*(volatile unsigned char*)0xfffff432)
#define PORT_G_SEL  (*(volatile unsigned char*)0xfffff433)

#define SPIM_CONT   (*(volatile unsigned short*)0xfffff802)
#define SPIM_DATA   (*(volatile unsigned short*)0xfffff800)

#define PORT_ISR    (*(volatile unsigned long*)0xfffff30c)
#define PORT_IMR    (*(volatile unsigned long*)0xFFFFF304)
#define PORT_IPR    (*(volatile unsigned long*)0xFFFFF310)
#define PORT_ICR    (*(volatile unsigned short*)0xFFFFF302)

#define PORT_CSA    (*(volatile unsigned short*)0xfffff110)
#define PORT_CSB    (*(volatile unsigned short*)0xfffff112)
#define PORT_CSC    (*(volatile unsigned short*)0xfffff114)
#define PORT_CSD    (*(volatile unsigned short*)0xfffff116)
#define PORT_CSGBA    (*(volatile unsigned short*)0xfffff100)
#define PORT_CSGBB    (*(volatile unsigned short*)0xfffff102)
#define PORT_CSGBC    (*(volatile unsigned short*)0xfffff104)
#define PORT_CSGBD    (*(volatile unsigned short*)0xfffff106)
#define PORT_PLLCR    (*(volatile unsigned short*)0xfffff200)
#define PORT_PLLFSR    (*(volatile unsigned short*)0xfffff202)
#define PORT_PCTLR    (*(volatile unsigned char*)0xfffff207)
#define PORT_DRAMMC    (*(volatile unsigned short*)0xfffffc00)
#define PORT_DRAMC    (*(volatile unsigned short*)0xfffffc02)
#define PORT_LCDCON    (*(volatile unsigned char*)0xfffffa27)
#define PORT_PWMR    (*(volatile unsigned short*)0xfffffa36)
#define PORT_TCTL    (*(volatile unsigned short*)0xfffff600)
#define PORT_TSTAT    (*(volatile unsigned short*)0xfffff60a)

#define PORT_MID    (*(volatile unsigned char*)0xfffff005)

#define PWMP (*(volatile unsigned char*)0xFFFFF504)
#define PWMS1 (*(volatile unsigned short*)0xFFFFF502)
#define PWMC (*(volatile unsigned short*)0xFFFFF500)

#define RTCTIME (*(volatile unsigned long*)0xFFFFFb00)
#define UART (*(volatile unsigned long*)0xFFFFF900)


#define WAIT_STABLE	5


enum mc68328digi_mode {
	MC68328DIGI_RARE,		/* debounced up/down and absolute x,y */
	MC68328DIGI_DEBUG,		/* up/down, debounced, transitions, button */
	MC68328DIGI_PS2,		/* ps2 relative (default) */ 
};


struct mc68328digi_params {
	enum mc68328digi_mode mode;
	int	bounce_interval;
	int	sticky_drag;
	int	tap_interval;
	int	irq;
	int	io;
	int	calibrated;
	long	x_a, x_b;
	long	y_a, y_b;
};

#define MS *HZ/1000

/* Appears as device major=10 (MISC), minor=PC110_PAD */
#define MC68328DIGI_MINOR		9
#define ADS7843_DIGI_MAJOR		127

#define MC68328DIGI_IOCTL_TYPE		0x9a

#define MC68328DIGIIOCGETP _IOR(MC68328DIGI_IOCTL_TYPE, 0, struct mc68328digi_params)
#define MC68328DIGIIOCSETP _IOR(MC68328DIGI_IOCTL_TYPE, 1, struct mc68328digi_params)

/********** Interrupt number definitions ************/
#define DIGI_IRQ	20
#define KB_INT0	8
#define KB_INT1	9
#define KB_INT2	10
#define KB_INT3	11
#define KB_MINT	6
#define HOTSYNC_INT	16

/********** Device number ************/
#define KEYBOARD_MAJOR	126


#define MAX_BATTERY_VALUE	0xd7
#define MIN_BATTERY_VALUE	0xa8

#define MAX_CONTRAST_VALUE	0xc8
#define MIN_CONTRAST_VALUE	0x9c

 
#endif /* _MC68328DIGI_H */
