/* mc68328digi.c
 *
 * Linux driver to make the PalmPilot's touchpad behave
 * like a PS/2 mouse.
 *
 * My apologies, this is currently a complete mess. It'll notice taps
 * on Copilot and a real Pilot, but can't figure out the coordinates on the Pilot.
 *
 * Based on touchpad driver.
 * Copyright (C) 1998  Kenneth Albanowski <kjahds@kjahds.com>,
 *                     The Silver Hammer Group, Ltd.
 *
 * Original History
 * -------
 *	0.0 1997-05-16 Alan Cox <alan@cymru.net> - Pad reader
 *	0.1 1997-05-19 Robin O'Leary <robin@acm.org> - PS/2 emulation
 *	0.2 1997-06-03 Robin O'Leary <robin@acm.org> - tap gesture
 *	0.3 1997-07-10 Robin O'Leary <robin@acm.org> - sticky drag
 *
 * Distribution
 * ------------
 * Copyright (C) 1997-06-18 Robin O'Leary <robin@acm.org>  All rights reserved.
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Get the latest version from ftp://swan.ml.org/pub/pc110/
 */
 

#include <linux/config.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/kernel.h>
#include <linux/busmouse.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <linux/timer.h>


#include "digi_mc68328.h"


static struct mc68328digi_params current_params;


/* driver/filesystem interface management */
//static int active=0;	/* number of concurrent open()s */

static short active = 0;

extern short digi_data_type;	// digitizer data format
/*
 * Utility to reset a timer to go off some time in the future.
 */



void digi_interrupt(int irq, void *ptr, struct pt_regs *regs)
{
	unsigned long flags;
	int new_x, new_y, new_down;
	volatile int i;
	
	disable_pen_irq();
	//save_flags(flags); cli();
	
	#define PFDATA (*((volatile unsigned char*)0xFFFFF429))	

	(*(volatile unsigned char*)0xFFFFF44A) = 0xad; /* Switch off Port M Pen IRQ pullup */
	(*(volatile unsigned char*)0xFFFFF44B) = 0xd0;  /* Turn on Port M Pen IRQ select */
	
	PFDATA = (PFDATA & 0xf0) | 5;
	(*(volatile unsigned char*)0xFFFFF429) = 0xD9;    /* Set PFDATA to D9, Measure Y */
	(*(volatile unsigned short*)0xFFFFF802) = 0x427f; /* Set SPIMCONT to 427f */
	(*(volatile unsigned short*)0xFFFFF800) = 0xc000; /* Set SPIMDATA to c000, Preload data*/
	PFDATA &= ~0x80;

	(*((volatile unsigned short*)0xFFFFF802)) &= ~0x0100; /* Set SPIMCONT to 427f */
	(*((volatile unsigned short*)0xFFFFF802)) |= 0x0100; /* Set SPIMCONT to 437f */
	
	while(!((*((volatile unsigned short*)0xFFFFF802)) & 0x80))
		;

	new_y = (*((volatile unsigned short*)0xFFFFF800)); /* Read SPIMDATA */
	
	PFDATA |= 0x80;

	(*(volatile unsigned short*)0xFFFFF802) = 0x407f;/* Set SPIMCONT to 407f */
	
	(*(volatile unsigned char*)0xFFFFF429) = 0xD5;    /* Set PFDATA to D5 */
	(*(volatile unsigned char*)0xFFFFF429) = 0xD6;    /* Set PFDATA to D6 */
	(*(volatile unsigned short*)0xFFFFF802) = 0x427f;/* Set SPIMCONT to 427f */
	(*(volatile unsigned short*)0xFFFFF800) = 0xffff; /* Set SPIMDATA to ffff */

	PFDATA &= ~0x80;
	(*(volatile unsigned short*)0xFFFFF802) = 0x427f;/* Set SPIMCONT to 427f */
	(*((volatile unsigned short*)0xFFFFF802)) = 0x437f; /* Set SPIMCONT to 437f */

	while(!((*((volatile unsigned short*)0xFFFFF802)) & 0x80)) 
		;

	new_x = (*((volatile unsigned short*)0xFFFFF800)); /* Read SPIMDATA */

	PFDATA |= 0x80;

	(*(volatile unsigned short*)0xFFFFF802) = 0x407f;/* Set SPIMCONT to 407f */

	(*(volatile unsigned char*)0xFFFFF429) = 0xD5;    /* Set PFDATA to d5 */
	(*(volatile unsigned char*)0xFFFFF429) = 0xDD;    /* Set PFDATA to dd */
	
	(*(volatile unsigned char*)0xFFFFF44A) = 0xed; /* Switch off Port M Pen IRQ pullup */
	(*(volatile unsigned char*)0xFFFFF44B) = 0x90;  /* Turn on Port M Pen IRQ select */

	
	//enable_pen_irq();

	// In fact there is no message queue here, COZ the PC's response is too fast
	//if (currentMsg > 0) disable_pen_irq();	

	//restore_flags(flags);
	digi_timer_callback(1);
}

/*
 * close access to the pad
 */
static void close_digi(struct inode * inode, struct file * file)
{
	if (--active)
		return;

	(*((volatile unsigned long*)0xFFFFF304)) |= (1<<20);   /* Enable interrupt */

	MOD_DEC_USE_COUNT;
}


/*
 * open access to the pad
 */
static int open_digi(struct inode * inode, struct file * file)
{
	if (active++)
		return 0;
	MOD_INC_USE_COUNT;
	return 0;
}

/*
 * writes are disallowed
 */
static int write_digi(struct inode * inode, struct file * file, const char * buffer, int count)
{
	return -EINVAL;
}

void enable_pen_irq(void)
{
	/* Enable Pen interrupt */
	(*((volatile unsigned long*)0xFFFFF304)) &= ~(1<<20);   /* Enable interrupt */
}
void disable_pen_irq(void)
{
	/* Disable Pen interrupt */
	(*((volatile unsigned long*)0xFFFFF304)) |= (1<<20);   /* Enable interrupt */
}

static struct file_operations pad_fops = {
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,open_digi,NULL,NULL,NULL,
};


static struct miscdevice mc68328_digi = {
	MC68328DIGI_MINOR, "mc68328 digitizer", &pad_fops
};


int mc68328digi_init(void)
{	
	if(M68EZ328_request_irq(20, digi_interrupt, 0, "mc68328digi", 0))
	{
		printk("mc68328digi: Unable to get IRQ.\n");
		return -EBUSY;
	}


	if (register_chrdev(127,"digi",&pad_fops) != 0) printk("register digitizer error\n");

	(*(volatile unsigned char*)0xFFFFF42B) = 0xFF; /*Setting PFSEL byte to ff*/
	(*(volatile unsigned char*)0xFFFFF428) = 0xFF; /*Setting PFDIR byte to ff*/
	(*(volatile unsigned char*)0xFFFFF42A) = 0x00; /*Setting PFPUEN byte to 0*/

	(*(volatile unsigned char*)0xFFFFF429) = 0xc5; /*Setting PFDATA byte to c5*/
	(*(volatile unsigned char*)0xFFFFF429) = 0xd5; /*Setting PFDATA byte to d5*/
	(*(volatile unsigned char*)0xFFFFF448) = 0x50; /*Setting PMDIR byte to 50*/
	(*(volatile unsigned char*)0xFFFFF449) = 0x00; /*Setting PMDATA byte to 0*/
	(*(volatile unsigned char*)0xFFFFF44B) = 0xd0; /*Setting PMSEL byte to d0*/
	(*(volatile unsigned char*)0xFFFFF429) = 0xd5; /*Setting PFDATA byte to d5*/

	(*(volatile unsigned short*)0xFFFFF802) = 0x427f; /*Setting SPIMCONT word to 427f*/
	(*(volatile unsigned short*)0xFFFFF800) = 0xffff; /*Setting SPIMDATA word to ffff*/
	(*(volatile unsigned char*)0xFFFFF429) = 0x55; /*Setting PFDATA byte to 55*/
	(*(volatile unsigned short*)0xFFFFF802) = 0x427f; /*Setting SPIMCONT word to 427f*/
	(*(volatile unsigned short*)0xFFFFF802) = 0x437f; /*Setting SPIMCONT word to 437f*/
	(void)(*(volatile unsigned short*)0xFFFFF800); /*Returning word from SPIMDATA ffff*/
	(*(volatile unsigned char*)0xFFFFF429) = 0xd5; /*Setting PFDATA byte to d5*/
	(*(volatile unsigned short*)0xFFFFF802) = 0x407f; /*Setting SPIMCONT word to 407f*/

	(*(volatile unsigned char*)0xFFFFF448) = 0x50; /*Setting PMDIR byte to 50*/
	(*(volatile unsigned char*)0xFFFFF449) = 0x00; /*Setting PMDATA byte to 0*/
	(*(volatile unsigned char*)0xFFFFF44A) = 0xED; /* Port M Pullups (dangerous?) */
	(*(volatile unsigned char*)0xFFFFF44B) = 0xd0; /*Setting PMSEL byte to d0*/
	(*(volatile unsigned char*)0xFFFFF429) = 0xd5; /*Setting PFDATA byte to d5*/
	(*(volatile unsigned char*)0xFFFFF429) = 0xdd; /*Setting PFDATA byte to dd*/

	
	(*((volatile unsigned long*)0xFFFFF304)) &= ~(1<<20);   /* Enable interrupt */
	(*(volatile unsigned char*)0xFFFFF44A) |= 64; /* Turn on Port M Pen IRQ pullup */
	(*(volatile unsigned char*)0xFFFFF44B) &= ~64;  /* Switch off Port M Pen IRQ select */
	

	return 0;
}


static void mc68328digi_unload(void)
{
	free_irq(current_params.irq, 0);

	current_params.irq=0;
	unregister_chrdev(127,"digi");
	//deregister_chrdev(&mc68328_digi);
}


int init_module(void)
{
	return mc68328digi_init();
}

void cleanup_module(void)
{
	mc68328digi_unload();
}
