/****************************************************************************
 *
 *  Copyright (C) 2001, Empower Technologies Inc.
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file COPYING in the main directory of this archive
 * for more details.
 *
 ***************************************************************************/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/kernel.h>
#include <linux/busmouse.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>

#include "digi_ads7843.h"
#include "irda.h"

static void PenTransferDelay(unsigned short ticks);
static unsigned short PenDetectY(void);
static unsigned short PenDetectX(void);
static void XchangeData(unsigned char cmd,unsigned short spimcont);
static int open_digi(void);
static int write_digi(void);

static struct mc68328digi_params default_params = {
	MC68328DIGI_PS2,50 MS,800 MS,300 MS,20,0x15E0,0,1<<16,0,1<<16,0
};

static struct file_operations pad_fops = {
	NULL,NULL,write_digi,NULL,NULL,NULL,NULL,NULL,NULL,open_digi,NULL,
};

static struct file_operations kb_fops = {
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
};



static struct mc68328digi_params current_params;
static U8 powerFlag = 1;
static U32 last_time=0;
static U8 lightFlag = 0;
static u16 icount = 0;
static unsigned long pIMR;
static U16 pPWMC;
static U16 pUART;
static U16 pSPIM;


#define CMD_READ_X 0xd8
#define CMD_READ_Y 0x98


/*************************************************
 
	Xchange data , Transfer a 8 bit command 
	to SPIM & reveive a 12bit return value

**************************************************/
static U16 GetPenPos(U16 cmd)
{
	U8 hibyte,lobyte;
	U16 newValue;
	XchangeData(cmd,0x8207);
	hibyte = SPIM_DATA;
	PenTransferDelay(50);
	/* Get high byte */
	XchangeData(0,0x8207);
	PenTransferDelay(100);
	hibyte = SPIM_DATA;
	/* Get low byte */
	XchangeData(0,0x8207);
	PenTransferDelay(100);
	lobyte = SPIM_DATA;
	newValue = hibyte;
	newValue = (newValue<<5);
	newValue |= lobyte>>3;
	//printk("h:%x, l:%x ",hibyte,lobyte);
	return (newValue);
}

/************************************************************
	Get ADS7843 data, use Palm method
************************************************************/

static U16 GetADSValue(U16 cmd)
{
	U16 newValue,oldSPIMCONT;
	oldSPIMCONT = SPIM_CONT;
	EnableCS();//	PORT_G_DAT &= 0xdf;  //enable A/D
	if (cmd & 0x10) {
		XchangeData(cmd >>1,0x2246);
		XchangeData(cmd & 0x0001,0x2240);		
	}else {
		XchangeData(cmd,0x2247);
	}
	PenTransferDelay(WAIT_STABLE);
	PORT_E_DIR |= 0x01;
	PORT_E_DAT &= 0xfe;
	PORT_E_PUL &= 0xfe;
	PORT_E_SEL |= 0x01;
	XchangeData(0,0x224f);
	newValue = ((SPIM_DATA>>3) & 0xfff);
	PORT_E_SEL &= 0xfe;
	DisableCS();
	SPIM_CONT = oldSPIMCONT;
//	PORT_G_DAT |= 0x20;  //disable A/D
	PenTransferDelay(WAIT_STABLE);
	
	return (newValue);
}

static U16 PenDetectX(void)
{
	U16 newValue;
	PenTransferDelay(WAIT_STABLE);
	newValue = GetADSValue(CMD_READ_X);	// Send command to select CH1 using externel clock mode
	return (newValue);
}

static U16 PenDetectY(void)
{
	U16 newValue;
	PenTransferDelay(WAIT_STABLE);
	newValue = GetADSValue(CMD_READ_Y);	// Send command to select CHO using externel clock mode
	return (newValue); 
}

static EnableCS(void)
{
	SPIM_CONT |= 0x0200;	/* enable the peripheral master mode */
	//Enable A/D, Port G CS low
	PORT_G_DAT &= ~0x20;
	/* Software power down A/D converter */
}

static DisableCS(void)
{
	/* Software power down A/D converter */
//	XchangeData(0x80,0x8207);
	/* Set C/S inactive high ---- disable A/D */
	PORT_G_DAT |= 0x20;
	SPIM_CONT &= 0xfeff;	/* disable the peripheral master mode */
}

U8 digi_interrupt(int irq, void *ptr, struct pt_regs *regs)
{
	int sample, xPos, yPos;
	U8 CanEnablePenIrq = 1;
	int prevX, prevY;
	
	/* Disable Pen interrupt, this line MUST be active in the future! */
	disable_pen_irq();
	
	prevX = PenDetectX();
	prevY = PenDetectY();
	
	/* loop untill a stable position arrive */
	sample = 1;
	if ((prevX == 0) || (prevY == 0)) sample = 0;
	
	xPos = PenDetectX();
	yPos = PenDetectY();
	
	printk("(%d,%d)\n",xPos,yPos);	
	digi_timer_callback(1);
}

/****************************************************************************
 *
 *	Initialize the ADS Digitizer params 
 *
 ****************************************************************************/
static void init_digi_params(void)
{
	/* initialize ADS 7843 controler */
//	PORT_E_SEL &= 0xf8;	/* set bit 0-2 zero */
	PORT_F_PUL |= 0x02;	/* Port F bit 1 clear, let it be PEN IRQe*/
	PORT_F_DIR |= 0x02;	/* Port F pin 1 for output purpose */
	PORT_F_DAT &= 0xfd;	/* Port F pin 1 for input purpose */
	PORT_F_SEL &= 0xfd;	/* Port F bit 1 clear, let it be PEN IRQe*/
	GetADSValue(0x90);
	PORT_G_SEL |= 0x20;	/* Port G bit 5 pull up, and set as I/O mode*/
	PORT_G_DIR |= 0x20;	/* Port G pin 5 for output purpose */
	PORT_G_DAT |= 0x20;	/* Set C/S inactive high ---- disable A/D */
}


static void XchangeData(U8 cmd,U16 spimcont)
{
	SPIM_CONT = spimcont;
	SPIM_DATA = cmd;	//The value wanna transfer
	// Do exchange
	SPIM_CONT |= 0x0100;	/* start exchange data */
	PenTransferDelay(WAIT_STABLE);
	// Pen poll done
	while (SPIM_CONT & 0x100) {;}
}


static int open_digi(void)
{
	return 0;
}

static int write_digi(void)
{
	return 0;
}


//only delay some ticks
static void PenTransferDelay(U16 ticks)
{
	while (ticks != 0,ticks --);
}

void disable_pen_irq(void)
{
	PORT_IMR |= (1<<20);
}

void enable_pen_irq(void)
{
	/* enable IRQ 5 */
	PORT_IMR &= ~(1<<20);   /* Enable interrupt */
}

int mc68328digi_init(void)
{
	current_params = default_params;

	if(M68EZ328_request_irq(current_params.irq, digi_interrupt, 0, "mc68328digi", 0))
	{
		printk("mc68328digi: Unable to get IRQ.\n");
		return -EBUSY;
	}
	if(check_region(current_params.io, 4))
	{
		printk("mc68328digi: I/O area in use.\n");
		free_irq(current_params.irq,0);
		return -EBUSY;
	}

	init_digi_params();
	if (register_chrdev(ADS7843_DIGI_MAJOR,"digi",&pad_fops) != 0) printk("register digitizer error\n");
	enable_pen_irq();

	
	return 0;
}

void mc68328digi_unload(void)
{
}

int init_module(void)
{
	return mc68328digi_init();
}

void cleanup_module(void)
{
	mc68328digi_unload();
}
