#ifndef _IRDA_H
#define _IRDA_H

#include <linux/ioctl.h>


typedef unsigned char U8;
typedef unsigned short U16;
typedef unsigned long U32;
typedef unsigned long long U64;

#define MC68328_SERIAL_MAJOR	123
#define MC68328_IRDA_MAJOR	125

#define PORT_B_DAT  (*(volatile unsigned char*)0xfffff409)
#define PORT_C_DAT  (*(volatile unsigned char*)0xfffff411)
#define PORT_D_DAT  (*(volatile unsigned char*)0xfffff419)
#define PORT_E_DAT  (*(volatile unsigned char*)0xfffff421)
#define PORT_F_DAT  (*(volatile unsigned char*)0xfffff429)
#define PORT_G_DAT  (*(volatile unsigned char*)0xfffff431)

#define PORT_B_DIR  (*(volatile unsigned char*)0xfffff408)
#define PORT_C_DIR  (*(volatile unsigned char*)0xfffff410)
#define PORT_D_DIR  (*(volatile unsigned char*)0xfffff418)
#define PORT_E_DIR  (*(volatile unsigned char*)0xfffff420)
#define PORT_F_DIR  (*(volatile unsigned char*)0xfffff428)
#define PORT_G_DIR  (*(volatile unsigned char*)0xfffff430)

#define PORT_B_PUL  (*(volatile unsigned char*)0xfffff40a)
#define PORT_C_PUL  (*(volatile unsigned char*)0xfffff412)
#define PORT_D_PUL  (*(volatile unsigned char*)0xfffff41a)
#define PORT_E_PUL  (*(volatile unsigned char*)0xfffff422)
#define PORT_F_PUL  (*(volatile unsigned char*)0xfffff42a)
#define PORT_G_PUL  (*(volatile unsigned char*)0xfffff432)

#define PORT_B_SEL  (*(volatile unsigned char*)0xfffff40b)
#define PORT_C_SEL  (*(volatile unsigned char*)0xfffff413)
#define PORT_D_SEL  (*(volatile unsigned char*)0xfffff41b)
#define PORT_E_SEL  (*(volatile unsigned char*)0xfffff423)
#define PORT_F_SEL  (*(volatile unsigned char*)0xfffff42b)
#define PORT_G_SEL  (*(volatile unsigned char*)0xfffff433)

#define PORT_ISR    (*(volatile unsigned long*)0xfffff30c)
#define PORT_IMR    (*(volatile unsigned long*)0xFFFFF304)
#define PORT_IPR    (*(volatile unsigned long*)0xFFFFF310)
#define PORT_ICR    (*(volatile unsigned short*)0xFFFFF302)

#define PORT_PLLFSR    (*(volatile unsigned short*)0xfffff202)
#define PORT_RTCCTL    (*(volatile unsigned short*)0xfffffB0C)
#define RTCCTL_XTL_MASK 0x0020

#define PORT_RTCTIME (*(volatile unsigned long*)0xFFFFFb00)
#define PORT_USTCNT (*(volatile unsigned short*)0xFFFFF900)
#define PORT_UBAUD (*(volatile unsigned short*)0xFFFFF902)
#define PORT_UMISC (*(volatile unsigned short*)0xFFFFF908)
#define PORT_NIPR (*(volatile unsigned short*)0xFFFFF90A)
#define PORT_URX (*(volatile unsigned char*)0xFFFFF904)
#define PORT_UTX (*(volatile unsigned char*)0xFFFFF906)
#define PORT_URXDATA (*(volatile unsigned char*)0xFFFFF905)
#define PORT_UTXDATA (*(volatile unsigned char*)0xFFFFF907)


#define WAIT_STABLE	10

#define PPPINITFCS 0xffff /* Initial FCS value */
#define PPPGOODFCS 0xf0b8 /* Good final FCS value */


#define UART_CON_PEN 0x0800		//default is no parity, set to make parity
#define UART_CON_ODD 0x0400		//default is even parity, set to make odd parity
#define UART_CON_STOP2 0x0200   //default is 1 stop-bit, set to make 2 stop-bit
#define UART_CON_7CHAR 0x0100	//default is 8-char transmision, set to make 7-char transmission 
#define UART_CON_REVERSE 0x000C  //default is polarity not reversed, set to make reversed


#define IRDA_LOF 0xFF
#define IRDA_BOF 0xC0
#define IRDA_EOF 0xC1
#define IRDA_CE 0x7D
#define IRDA_FRAME_START 0x01
#define IRDA_FRAME_INCONTROL 0x02
#define IRDA_FRAME_INFRAME 0x04
#define IRDA_FRAME_CLOSE 0x08


#define UART_TIME_OUT -1
#define UART_OPERATE_OK	0x00
#define IRDA_FCS_ERROR	-1
#define IRDA_SEND_ERROR	-1
#define IRDA_SEND_TOO_BIG	-1


#define UART_TIMEOUT_DELAY	15
#define UART_BUFFER_SIZE 0x7ff
#define UART_INT 2 
#endif /* _IRDA_H */
