/* httpd.c:  A very simple http server
 *
 * Copyright (C) 1998  Kenneth Albanowski <kjahds@kjahds.com>,
 *         1997, 1998  D. Jeff Dionne <jeff@ryeham.ee.ryerson.ca>,
 *                     The Silver Hammer Group, Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>
#include <sys/stat.h>
#include <dirent.h>
#include <signal.h>
#include <unistd.h>
#include <ctype.h>

#include "stamp.h"

int TIMEOUT=30;

#ifndef O_BINARY
#define O_BINARY 0
#endif

char referrer[128];
int content_length;

#define SERVER_PORT 80

int
PrintTextHeader(FILE *f)
{
  alarm(TIMEOUT);
  fprintf(f,"HTTP/1.0 200 OK\n");
  fprintf(f,"Content-type: text/plain\n");
  /*fprintf(f,"Date: Sun, 26 Apr 1998 12:00:00 GMT\n");*/
  fprintf(f,"Server: ipserver-httpd 0.0.1\n");
  fprintf(f,"Expires: 0\n");
  fprintf(f,"\n");
  alarm(0);
  return 0;
}

int
PrintHTMLHeader(FILE *f)
{
  alarm(TIMEOUT);
  fprintf(f,"HTTP/1.0 200 OK\n");
  fprintf(f,"Content-type: text/html\n");
  /*fprintf(f,"Date: Sun, 26 Apr 1998 12:00:00 GMT\n");*/
  fprintf(f,"Server: ipserver-httpd 0.0.1\n");
  fprintf(f,"Expires: 0\n");
  fprintf(f,"\n");
  alarm(0);
  return 0;
}

int
PrintGIFHeader(FILE *f)
{
  alarm(TIMEOUT);
  fprintf(f,"HTTP/1.0 200 OK\n");
  fprintf(f,"Content-type: image/gif\n");
  /*fprintf(f,"Date: Sun, 26 Apr 1998 12:00:00 GMT\n");*/
  fprintf(f,"Server: ipserver-httpd 0.0.1\n");
  fprintf(f,"Expires: 0\n");
  /*fprintf(f,"Expires: Fri, 31 Dec 1999 12:00:00 GMT\n");*/
  fprintf(f,"\n");
  alarm(0);
  return 0;
}

int
DoGif(FILE *f, char *name)
{
  char *buf;
  FILE * infile;
  int count;

  if (!(infile = fopen(name, "r"))) {
    alarm(TIMEOUT);
    fprintf(f, "Unable to open GIF file %s, %d\n", name, errno);
    fflush(f);
    alarm(0);
    return -1;
  }
  
  PrintGIFHeader(f);

  copy(infile,f); /* prints the page */  

  alarm(TIMEOUT);
  fclose(infile);
  alarm(0);
  
  return 0;
}

int
DoDir(FILE *f, char *name)
{
  char *buf;
  DIR * dir;
  struct dirent * dirent;

  if ((dir = opendir(name))== 0) {
    fprintf(f, "Unable to open directory %s, %d\n", name, errno);
    fflush(f);
    return -1;
  }
  
  PrintHTMLHeader(f);
  
  alarm(TIMEOUT);
  fprintf(f, "<H1>Index of %s</H1>\n\n",name);
  alarm(0);
  
  while(dirent = readdir(dir)) {
	alarm(TIMEOUT);
#if 0
  	if (dirent->d_name[0] == '.')
  		continue;
#endif
  	if (name)
  		fprintf(f, "<p><a href=\"/...%s/%s\">%s</a></p>\n", name/*+2*/, dirent->d_name, dirent->d_name);
  	else
  		fprintf(f, "<p><a href=\"/.../%s\">%s</a></p>\n", dirent->d_name, dirent->d_name);
	alarm(0);
  }
  
  closedir(dir);
  return 0;
}

int
DoHTML(FILE *f, char *name)
{
  char *buf;
  FILE *infile;
  int count;
  char * dir = 0;

  if (!(infile = fopen(name,"r"))) {
    alarm(TIMEOUT);
    fprintf(f, "Unable to open HTML file %s, %d\n", name, errno);
    fflush(f);
    alarm(0);
    return -1;
  }

  PrintHTMLHeader(f);
  
  if (dir = strrchr(name, '/')) {
  	*dir = '\0';
  	chdir(name);
  }

  copy(infile,f); /* prints the page */  
  
  if (dir) {
  	chdir ("/htdocs");
  }

  alarm(TIMEOUT);
  fclose(infile);
  alarm(0);

  return 0;
}

int
DoText(FILE *f, char *name)
{
  char *buf;
  FILE *infile;
  int count;

  if (!(infile = fopen(name,"r"))) {
    alarm(TIMEOUT);
    fprintf(f, "Unable to open HTML file %s, %d\n", name, errno);
    fflush(f);
    alarm(0);
    return -1;
  }

  PrintTextHeader(f);
  copy(infile,f); /* prints the page */  

  alarm(TIMEOUT);
  fclose(infile);
  alarm(0);

  return 0;
}

int
ParseReq(FILE *f, char *r)
{
  char *bp;
  struct stat stbuf;
  char * arg;
  char * c;
  int e;
  int raw;

#ifdef DEBUG
  printf("req is '%s'\n", r);
#endif
  
  while(*(++r) != ' ');  /*skip non-white space*/
  while(isspace(*r))
  	r++;
  
  while (*r == '/')
  	r++;

  bp = r;
  /**r = '.';*/
  
  while(*r && (*(r) != ' ') && (*(r) != '?'))
  	r++;
  	
#ifdef DEBUG
  printf("bp='%s', r='%s'\n", bp, r);
#endif
  	
  if (*r == '?') {
  	char * e;
  	*r = 0;
  	arg = r+1;
  	if (e = strchr(arg,' ')) {
  		*e = '\0';
  	}
  } else {
  	arg = 0;
  	*r = 0;
  }
  
  c = bp;

  if (c && !stat(c, &stbuf)) {
    if (S_ISDIR(stbuf.st_mode)) { 
    	char * end = c + strlen(c);
    	strcat(c, "/index.html");
    	if (!stat(c, &stbuf)) {
    		DoHTML(f, c);
    	} else {
    		*end = '\0';
	      DoDir(f,c);
	}
    }
    else if (!strcmp(r - 4, ".gif"))
      DoGif(f,c);
    else if (!strcmp(r - 5, ".html"))
      DoHTML(f,c);
    else
      DoText(f,c);
  } else {
  	int e = errno;
  	PrintHTMLHeader(f);
  	alarm(TIMEOUT);
  	fprintf(f, "<p>File '%s' not found (error %d)</p>\n",
  		bp, e);
  	alarm(0);
  }
  return 0;
}

void sigalrm(int signo)
{
	/* got an alarm, exit & recycle */
	exit(0);
}

int
main(int argc, char *argv[])
{
  int fd, s;
  int len;
  volatile int true = 1;
  struct sockaddr_in ec;
  struct sockaddr_in server_sockaddr;
  FILE *f;

  char buf[160];
  char buf1[160];

  check_version_argument;
          
  signal(SIGCHLD, SIG_IGN);
  signal(SIGPIPE, SIG_IGN);
  signal(SIGALRM, sigalrm);

  chdir("/htdocs");

restart:

  if((s = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) {
    perror("Unable to obtain network");
    exit(1);
  }
  
  if((setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (void *)&true, 
     sizeof(true))) == -1) {
    perror("setsockopt failed");
    exit(1);
  }

  server_sockaddr.sin_family = AF_INET;
  server_sockaddr.sin_port = htons(SERVER_PORT);
  server_sockaddr.sin_addr.s_addr = htonl(INADDR_ANY);

  if(bind(s, (struct sockaddr *)&server_sockaddr, 
     sizeof(server_sockaddr)) == -1)  {
    if ((errno == EADDRINUSE) && (getppid() != 1)) {
    	/* Someone launched httpd from the shell, when it was already running */
	printf("Please do not use this command\n");
    } else
    	perror("Unable to bind socket");
    exit(1);
  }

  if(listen(s, 8*3) == -1) { /* Arbitrary, 8 files/page, 3 clients */
    perror("Unable to listen");
    exit(4);
    goto restart;
  }

  while (1) {
    len = sizeof(ec);
    
    if((fd = accept(s, (void *)&ec, &len)) == -1) {
      exit(5);
      close(s);
      goto restart;
    }

    f = fdopen(fd,"a+");
    if (!f) {
       fprintf(stderr, "httpd: Unable to open httpd input fd, error %d\n", errno);
       alarm(TIMEOUT);
       close(fd);
       alarm(0);
       continue;
    }
    setbuf(f, 0);

    alarm(TIMEOUT);

    if (!fgets(buf, 150, f)) {
       fprintf(stderr, "httpd: Error reading connection, error %d\n", errno);
       fclose(f);
       alarm(0);
       continue;
    }
#ifdef DEBUG
    printf("buf = '%s'\n", buf);
#endif
    
    alarm(0);

    referrer[0] = '\0';
    content_length = -1;
    
    alarm(TIMEOUT);
    while (fgets(buf1, 150, f) && (strlen(buf1) > 2)) {
	alarm(TIMEOUT);
#ifdef DEBUG
    	printf("Got buf1 '%s'\n", buf1);
#endif
	if (!strncasecmp(buf1, "Referer:", 8)) {
		char * c = buf1+8;
    		while (isspace(*c))
    			c++;
    		strcpy(referrer, c);
    	} else if (!strncasecmp(buf1, "Referrer:", 9)) {
		char * c = buf1+9;
    		while (isspace(*c))
    			c++;
    		strcpy(referrer, c);
    	} else if (!strncasecmp(buf1, "Content-length:", 15)) {
    		content_length = atoi(buf1+15);
    	} 
    }
    alarm(0);


    if (ferror(f)) {
    	fprintf(stderr, "http: Error continuing reading connection, error %d\n", errno);
    	fclose(f);
    	continue;
    }

    
    ParseReq(f, buf);

    alarm(TIMEOUT);
    fflush(f);
    fclose(f);
    alarm(0);

  }
}
