/* gencrctable.c: Generate a compilable precomputed CRC table
 *
 * Copyright (C) 1998  Kenneth Albanowski <kjahds@kjahds.com>,
 *                     The Silver Hammer Group, Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>


static char * program;
static void usage(void)
{
	printf("Usage: %s <polynomial> <bits> <reflect>\n", program);
	printf("\nFor example, '%s 0x1021 16 1'\n", program);
	printf("\nBits must be 8, 16, or 32.\n");
	printf("\nReflect must be 1 or 0.\n");
	exit(0);
}

int main(int argc, char * argv[])
{
	int bits;
	unsigned long poly;
	int reflect;
	char * e;
	char * type;
	char tname[64], fname[64];
	int i, acc;
	FILE * f;
	
	program = argv[0];

	if (argc != 4) {
		usage();
	}
	
	poly = strtoul(argv[1], &e, 0);
	
	if (e == argv[1])
		usage();
	
	bits = strtoul(argv[2], &e, 0);
	
	if (e == argv[2])
		usage();
	
	reflect = !!strtoul(argv[3], &e, 0);
	
	if (e == argv[3])
		usage();
	
	switch (bits) {
	case 8:
		type = "unsigned char";
		sprintf(tname, "crc_0x%02X_%d", poly, reflect);
		break;
	case 16:
		type = "unsigned short int";
		sprintf(tname, "crc_0x%04X_%d", poly, reflect);
		break;
	case 32:
		type = "unsigned int";
		sprintf(tname, "crc_0x%08X_%d", poly, reflect);
		break;
	default:
		usage();
	}
	
	strcpy(fname, tname); 
	strcat(fname, ".c");
	f = fopen(fname, "w");
	
	fprintf(f, "\n%s %s[256] = {\n", type, tname);
	
	acc = 0;
	for (i=0; i<256; i++) {
		int bit;
	        if (reflect) {
	        	acc = i;
	        	for (bit = 0; bit < 8; bit++)
	        		if (acc & 1)
	        			acc = (acc >> 1) ^ poly;
	        		else
	        			acc = (acc >> 1);
	        } else {
	        	acc = i << (bits-8);
	        	for (bit = 0; bit < 8; bit++)
	        		if (acc & (1 << (bits - 1)))
	        			acc = (acc << 1) ^ poly;
	        		else
	        			acc = (acc << 1);
	        }
	        switch (bits) {
	        case 8:
	        	fprintf(f, "\t0x%02X,\n", acc & 0xff);
	        	break;
	        case 16:
	        	fprintf(f, "\t0x%04X,\n", acc & 0xffff);
	        	break;
	        case 32:
	        	fprintf(f, "\t0x%08X,\n", acc & 0xffffffff);
	        	break;
	        }
	}
	fprintf(f, "};\n");
		
	fclose(f);
	
	printf("Please add %s to link\n", fname);
	
}
